#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import logging
from datetime import datetime
from typing import AsyncIterator, Union, List

import pyrogram
from pyrogram import raw, types, utils

log = logging.getLogger(__name__)


async def get_chunk(
    client: "pyrogram.Client",
    chat_id: Union[int, str],
    *,
    limit: int = 0,
    offset: int = 0,
    offset_id: int = 0,
    from_date: datetime = utils.zero_datetime(),
    min_id: int = 0,  # Inclusive
    max_id: int = 0,  # Inclusive
    reverse: bool = False,
) -> List["types.Message"]:
    # Telegram API requires `offset_id` as starting point, boundaries alone don't work
    if (min_id or max_id) and not offset_id:
        if max_id:
            offset_id = max_id + 1
        elif min_id:
            offset_id = 0

    if min_id and max_id and not offset_id:
        offset_id = max_id + 1

    history: raw.base.messages.Messages = await client.invoke(
        raw.functions.messages.GetHistory(
            peer=await client.resolve_peer(chat_id),
            offset_id=offset_id,
            offset_date=utils.datetime_to_timestamp(from_date),
            add_offset=offset,
            limit=limit,
            max_id=max_id,
            min_id=min_id,
            hash=0
        ),
        sleep_threshold=60
    )

    messages = await utils.parse_messages(client, history, replies=0)

    if reverse:
        messages.reverse()
        
    return messages


class GetChatHistory:
    async def get_chat_history(
        self: "pyrogram.Client",
        chat_id: Union[int, str],
        limit: int = 0,
        offset: int = 0,
        offset_id: int = None,
        offset_date: datetime = utils.zero_datetime(),
        min_id: int = 0,
        max_id: int = 0,
        reverse: bool = False,
    ) -> AsyncIterator["types.Message"]:
        """Get messages from a chat history.

        The messages are returned in reverse chronological order.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".
                For a contact that exists in your Telegram address book you can use his phone number (str).

            limit (``int``, *optional*):
                Limits the number of messages to be retrieved.
                By default, no limit is applied and all messages are returned.

            offset (``int``, *optional*):
                Sequential number of the first message to be returned.
                Negative values are also accepted and become useful in case you set offset_id or offset_date.

            offset_id (``int``, *optional*):
                Identifier of the first message to be returned.
                This parameter is deprecated and should not be used.
                Use `min_id` / `max_id` instead for proper filtering.

            offset_date (:py:obj:`~datetime.datetime`, *optional*):
                Pass a date as offset to retrieve only older messages starting from that date.

            min_id (``int``, *optional*):
                If a positive value was provided, the method will return only messages
                with IDs more than or equal to min_id (inclusive).

            max_id (``int``, *optional*):
                If a positive value was provided, the method will return only messages
                with IDs less than or equal to max_id (inclusive).

            reverse (``bool``, *optional*):
                Pass True to retrieve the messages from oldest to newest.

        Returns:
            ``Generator``: A generator yielding :obj:`~pyrogram.types.Message` objects.

        Example:
            .. code-block:: python

                async for message in app.get_chat_history(chat_id):
                    print(message.text)
        """
        if offset_id is not None:
            log.warning(
                "`offset_id` is deprecated and will be removed in future updates. " \
                "Use `min_id` or `max_id` instead."
            )

        current: int = 0
        total: int = limit or (1 << 31) - 1
        limit: int = min(100, total)  # We can only retrieve up to 100 messages at a time

        min_id = (min_id - 1) if min_id else 0  # Make `min_id` inclusive
        max_id = (max_id + 1) if max_id else 0  # Make `max_id` inclusive

        if reverse:
            offset_id = min_id if min_id else 1
            offset = offset - limit
        else:
            offset_id = max_id if max_id else 0

        while True:
            messages = await get_chunk(
                client=self,
                chat_id=chat_id,
                limit=limit,
                offset=offset,
                offset_id=offset_id,
                from_date=offset_date,
                max_id=max_id,
                min_id=min_id,
                reverse=reverse
            )
            
            if not messages:
                return

            offset_id = messages[-1].id + (1 if reverse else 0)
            for message in messages:
                yield message

                current += 1
                
                if current >= total:
                    return



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

from pyrogram import raw

from ..object import Object


class ChecklistTasksDone(Object):
    """Some tasks from a checklist were marked as done or not done.

    Parameters:
        checklist_message_id (``int``):
            Identifier of the message with the checklist.
            Can be None if the message was deleted.

        marked_as_done_task_ids (List of ``int``):
            Identifiers of tasks that were marked as done

        marked_as_not_done_task_ids (List of ``int``):
            Identifiers of tasks that were marked as not done
    """

    def __init__(
        self,
        *,
        checklist_message_id: int,
        marked_as_done_task_ids: List[int],
        marked_as_not_done_task_ids: List[int]
    ):

        super().__init__()

        self.checklist_message_id = checklist_message_id
        self.marked_as_done_task_ids = marked_as_done_task_ids
        self.marked_as_not_done_task_ids = marked_as_not_done_task_ids

    @staticmethod
    def _parse(message: "raw.types.MessageService") -> "ChecklistTasksDone":
        action: "raw.types.MessageActionTodoCompletions" = message.action

        return ChecklistTasksDone(
            checklist_message_id=getattr(message.reply_to, "reply_to_msg_id", None),
            marked_as_done_task_ids=action.completed,
            marked_as_not_done_task_ids=action.incompleted
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from pyrogram import raw
from ..object import Object


class ForumTopicCreated(Object):
    """A service message about a new forum topic created in the chat.


    Parameters:
        id (``int``):
            Id of the topic.

        title (``str``):
            Name of the topic.

        icon_color (``int``):
            Color of the topic icon in decimal format.

        custom_emoji_id (``int``, *optional*):
            Unique identifier of the custom emoji shown as the topic icon.
    """

    def __init__(
        self, *,
        id: int,
        title: str,
        icon_color: int,
        custom_emoji_id: int = None
    ):
        super().__init__()

        self.id = id
        self.title = title
        self.icon_color = icon_color
        self.custom_emoji_id = custom_emoji_id

    @staticmethod
    def _parse(message: "raw.base.Message") -> "ForumTopicCreated":
        return ForumTopicCreated(
            id=getattr(message, "id", None),
            title=getattr(message.action, "title", None),
            icon_color=getattr(message.action, "icon_color", None),
            custom_emoji_id=getattr(message.action, "icon_emoji_id", None)
        )



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

from pyrogram import raw
from ..object import Object


class Location(Object):
    """A point on the map.

    Parameters:
        longitude (``float``, *optional*):
            Longitude as defined by sender.

        latitude (``float``, *optional*):
            Latitude as defined by sender.

        accuracy_radius (``int``, *optional*):
            The estimated horizontal accuracy of the location, in meters as defined by the sender.

        address (``str``, *optional*):
            Textual description of the address (mandatory).

        live_period (``int``, *optional*):
            For live locations, the time relative to the message send date, for which the location can be updated, in seconds.

        heading (``int``, *optional*):
            For live locations, a direction in which the location moves, in degrees; 1-360.

        proximity_alert_radius (``int``, *optional*):
            For live locations, a maximum distance to another chat member for proximity alerts, in meters (0-100000).
    """

    def __init__(
        self,
        *,
        longitude: Optional[float] = None,
        latitude: Optional[float] = None,
        accuracy_radius: Optional[int] = None,
        address: Optional[str] = None,
        live_period: Optional[int] = None,
        heading: Optional[int] = None,
        proximity_alert_radius: Optional[int] = None
    ):
        super().__init__()

        self.longitude = longitude
        self.latitude = latitude
        self.accuracy_radius = accuracy_radius
        self.address = address
        self.live_period = live_period
        self.heading = heading
        self.proximity_alert_radius = proximity_alert_radius

    @staticmethod
    def _parse(geo_point: "raw.types.GeoPoint") -> Optional["Location"]:
        if isinstance(geo_point, raw.types.GeoPoint):
            return Location(
                longitude=geo_point.long,
                latitude=geo_point.lat,
                accuracy_radius=geo_point.accuracy_radius,
            )

    @staticmethod
    def _parse_business(location: "raw.types.BusinessLocation") -> "Location":
        if isinstance(location, raw.types.BusinessLocation):
            longitude = None
            latitude = None
            accuracy_radius = None

            if isinstance(location.geo_point, raw.types.GeoPoint):
                longitude = location.geo_point.long
                latitude = location.geo_point.lat
                accuracy_radius = location.geo_point.accuracy_radius

            return Location(
                longitude=longitude,
                latitude=latitude,
                accuracy_radius=accuracy_radius,
                address=location.address
            )

    @staticmethod
    def _parse_media(media: "raw.types.MessageMediaGeoLive") -> Optional["Location"]:
        if isinstance(media, raw.types.MessageMediaGeoLive):
            parsed_location = Location._parse(media.geo)

            parsed_location.live_period = media.period
            parsed_location.heading = media.heading
            parsed_location.proximity_alert_radius = media.proximity_notification_radius

            return parsed_location



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import List

import pyrogram
from pyrogram import raw
from pyrogram import types
from pyrogram.errors import PhoneMigrate, NetworkMigrate

log = logging.getLogger(__name__)


class SendCode:
    async def send_code(
        self: "pyrogram.Client",
        phone_number: str,
        current_number: bool = None,
        allow_flashcall: bool = None,
        allow_app_hash: bool = None,
        allow_missed_call: bool = None,
        allow_firebase: bool = None,
        logout_tokens: List[bytes] = None,
        token: str = None,
        recaptcha_token: str = None,
        app_sandbox: bool = None,
    ) -> "types.SentCode":
        """Send the confirmation code to the given phone number.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            phone_number (``str``):
                Phone number in international format (includes the country prefix).

            current_number (``bool``, *optional*):
                Whether the phone number is the current one.

            allow_flashcall (``bool``, *optional*):
                Whether to allow a flash call.

            allow_app_hash (``bool``, *optional*):
                Whether to allow an app hash.

            allow_missed_call (``bool``, *optional*):
                Whether to allow a missed call.

            allow_firebase (``bool``, *optional*):
                Whether to allow firebase.

            logout_tokens (List of ``bytes``, *optional*):
                List of logout tokens.

            token (``str``, *optional*):
                Token.

            recaptcha_token (``str``, *optional*):
                Recaptcha token.

            app_sandbox (``bool``, *optional*):
                Whether to use the app sandbox.

        Returns:
            :obj:`~pyrogram.types.SentCode`: On success, an object containing information on the sent confirmation code
            is returned.

        Raises:
            BadRequest: In case the phone number is invalid.
        """
        phone_number = phone_number.strip(" +")

        while True:
            try:
                r = await self.invoke(
                    raw.functions.auth.SendCode(
                        phone_number=phone_number,
                        api_id=self.api_id,
                        api_hash=self.api_hash,
                        settings=raw.types.CodeSettings(
                            allow_flashcall=allow_flashcall,
                            current_number=current_number,
                            allow_app_hash=allow_app_hash,
                            allow_missed_call=allow_missed_call,
                            allow_firebase=allow_firebase,
                            logout_tokens=logout_tokens,
                            token=token,
                            app_sandbox=app_sandbox
                        )
                    ),
                    recaptcha_token=recaptcha_token
                )
            except (PhoneMigrate, NetworkMigrate) as e:
                dc_option = await self.get_dc_option(e.value, ipv6=self.ipv6)
                await self.session.stop()

                self.session = await self.get_session(
                    dc_id=e.value,
                    server_address=dc_option.ip_address,
                    port=dc_option.port,
                    export_authorization=False,
                    temporary=True
                )

                await self.storage.dc_id(e.value)
                await self.storage.server_address(dc_option.ip_address)
                await self.storage.port(dc_option.port)
                await self.storage.auth_key(self.session.auth_key)
            else:
                return types.SentCode._parse(r)



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import List
import pyrogram
from pyrogram import types, enums


class GetChatsForFolderInviteLink:
    async def get_chats_for_folder_invite_link(
        self: "pyrogram.Client",
        chat_folder_id: int
    ) -> List["types.Chat"]:
        """Returns chats from a chat folder, suitable for adding to a chat folder invite link.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_folder_id (``int``):
                Unique identifier (int) of the target folder.

        Returns:
            List of :obj:`~pyrogram.types.Chat`: On success, list of suitable chats is returned.

        Example:
            .. code-block:: python

                for chat in await app.get_chats_for_folder_invite_link(folder_id):
                    print(chat.title)
        """
        folder = None
        chats = types.List()

        for f in await self.get_folders():
            if f.id == chat_folder_id:
                folder = f
                break

        if not folder:
            raise ValueError("Folder not found")

        if (
            folder.excluded_chats
            or folder.exclude_muted
            or folder.exclude_read
            or folder.exclude_archived
            or folder.include_contacts
            or folder.include_non_contacts
            or folder.include_bots
            or folder.include_groups
            or folder.include_channels
        ):
            return chats

        pinned_chats = folder.pinned_chats or []
        included_chats = folder.included_chats or []
        available_chats = pinned_chats + included_chats

        for chat in available_chats:
            if chat.type in (enums.ChatType.FORUM, enums.ChatType.SUPERGROUP, enums.ChatType.CHANNEL):
                chats.append(chat)

        return chats



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union

import pyrogram
from pyrogram import enums, raw, types, utils


class OpenWebApp:
    async def open_web_app(
        self: "pyrogram.Client",
        chat_id: Union[int, str],
        bot_user_id: Union[int, str],
        url: Optional[str] = None,
        message_thread_id: Optional[int] = None,
        direct_messages_topic_id: Optional[int] = None,
        reply_parameters: Optional["types.ReplyParameters"] = None,
        platform: Optional["enums.ClientPlatform"] = None
    ) -> str:
        """Informs pyrogram that a Web App is being opened from the attachment menu,
        a :obj:`~pyrogram.types.MenuButton`, an url,
        or an :obj:`~pyrogram.types.InlineKeyboardButton` button.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            chat_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.

            bot_user_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target bot.

            url (``str``, *optional*):
                The URL from an :obj:`~pyrogram.types.InlineKeyboardButton` or a :obj:`~pyrogram.types.MenuButton`.

            message_thread_id (``int``, *optional*):
                If not None, the message thread identifier to which the message will be sent.

            direct_messages_topic_id (``int``, *optional*):
                If not None, unique identifier of the topic of channel direct messages chat to which the message will be sent.

            reply_parameters (:obj:`~pyrogram.types.ReplyParameters`, *optional*):
                Information about the message or story to be replied in the message sent by the Web App.

            platform (:obj:`~pyrogram.enums.ClientPlatform`, *optional*):
                The platform on which the link will be opened.

        Returns:
            ``str``: On success, returns the url of a Web App.

        Example:
            .. code-block:: python

                link = await app.open_web_app(chat_id, bot_user_id)
        """
        if platform is None:
            platform = self.client_platform

        r = await self.invoke(
            raw.functions.messages.RequestWebView(
                peer=await self.resolve_peer(chat_id),
                bot=await self.resolve_peer(bot_user_id),
                platform=platform.value,
                from_bot_menu=True if url is None else None,
                url=url,
                reply_to=await utils.get_reply_to(
                    client=self,
                    reply_parameters=reply_parameters,
                    message_thread_id=message_thread_id,
                    direct_messages_topic_id=direct_messages_topic_id
                ),

            )
        )

        return r.url



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Union

import pyrogram
from pyrogram import raw, types, utils


class CreateGiftCollection:
    async def create_gift_collection(
        self: "pyrogram.Client",
        owner_id: Union[int, str],
        name: str,
        gift_ids: List[str]
    ) -> "types.GiftCollection":
        """Creates a collection from gifts on the current user's or a channel's profile page.

        .. include:: /_includes/usable-by/users.rst

        Parameters:
            owner_id (``int`` | ``str``):
                Unique identifier (int) or username (str) of the target chat.
                For your personal cloud (Saved Messages) you can simply use "me" or "self".

            name (``str``):
                Name of the collection, 1-12 characters.

            gift_ids (List of ``str``):
                Identifier of the gifts to add to the collection.

        Returns:
            :obj:`~pyrogram.types.GiftCollection`: On success, a created collection is returned.

        Example:
            .. code-block:: python

                await create_gift_collection("me", "My best gifts!", ["https://t.me/nft/NekoHelmet-9215"])
        """
        r = await self.invoke(
            raw.functions.payments.CreateStarGiftCollection(
                peer=await self.resolve_peer(owner_id),
                title=name,
                stargift=[await utils.get_input_stargift(self, owned_gift_id) for owned_gift_id in gift_ids],
            )
        )

        return await types.GiftCollection._parse(self, r)
